import glob
import math
import os
import numpy as np
from astropy.io import fits
from . import cosutil
from . import ccos

NOT_APPLICABLE = "N/A"
SEC_PER_DAY = 86400.            # seconds in a day

def splittag(infiles, outroot, starttime=None, increment=None, endtime=None,
             time_list=None, verbosity=1):
    """Split TIME-TAG files into multiple files.

    All times are in seconds, and the zero point is EXPSTART.

    Parameters
    ----------
    infiles: str
        Name of input file, possibly including a wildcard character.

    outroot: str
        Root name with which to construct output file names.

    starttime: float or None
        Time at beginning of first interval, or None if time_list was
        specified.

    increment: float or None
        Length of each time interval, or None if time_list was specified.

    endtime: float or None
        Time at end of last interval, or None if time_list is specified.

    time_list: str
        String containing explicit times of beginning of each time
        interval and end of the last interval; the first time may be given
        as "start" (time=0, i.e. at EXPSTART), and the last time may be
        given as either "stop" or "end" (time of last event in TIME column).

    verbosity: int {0, 1, 2}
        0 --> print almost nothing, 1 --> print some info,
        2 --> print more info
    """

    infiles = os.path.expandvars(infiles)
    outroot = os.path.expandvars(outroot)
    inlist = glob.glob(infiles)
    for input in inlist:
        splitOneTag(input, outroot, starttime, increment, endtime,
                    time_list, verbosity)

def splitOneTag(input, outroot, starttime=None, increment=None, endtime=None,
                time_list=None, verbosity=1):
    """Split a TIME-TAG file into multiple files.

    Parameters
    ----------
    input: str
        Name of input file

    outroot: str
        Root name with which to construct output file names

    starttime: float or None
        Time at beginning of first interval, or None

    increment: float or None
        Length of each time interval, or None

    endtime: float or None
        Time at end of last interval, or None

    time_list: str
        String containing explicit times of beginning of each
        time interval and end of the last interval

    verbosity: int {0, 1, 2}
        Indicates how much should be printed
    """

    cosutil.setVerbosity(verbosity)

    if not cosutil.isCorrtag(input):
        raise RuntimeError("%s is not a corrtag file" % input)

    (inroot, suffix) = splitName(input)

    ifd = fits.open(input, mode="copyonwrite")
    phdr = ifd[0].header
    try:
        hdr = ifd[("events")].header
    except KeyError:
        ifd.close()
        raise RuntimeError("%s is not a corrtag file" % input)
    data = ifd[("events")].data
    time_col = data.field("TIME").astype(np.float64)

    info = getInfo(input, phdr, hdr)
    if info["wavecorr"] != "COMPLETE":
        cosutil.printWarning("WAVECORR was not done for " + input)
    gti_hdu = getGTI(ifd)
    timeline_hdu = getTimeline(ifd)
    time_list = convertToSlices(time_col,
                                starttime, increment, endtime, time_list)
    cosutil.printMsg("time_list = %s" % repr(time_list), 2)

    # define output columns based on input table
    cd = ifd[("events")].columns

    file_index = 1              # one indexing for output file names
    for (t0, t1) in time_list:

        (i, j) = determineSlice(time_col, t0, t1)
        nrows = j - i
        if nrows <= 0:
            cosutil.printWarning("no rows in increment %.2f to %.2f" %
                                 (t0, t1))
            continue

        filename = constructOutputName(outroot, file_index, suffix)
        ofd = fits.HDUList(fits.PrimaryHDU(header=phdr))
        hdu = fits.BinTableHDU.from_columns(cd, header=hdr, nrows=nrows)
        ofd.append(hdu)

        copyRows(data, ofd, i, j)
        nevents = j - i

        out_gti_hdu = createNewGTI(gti_hdu, t0, t1)
        if out_gti_hdu is not None:
            ofd.append(out_gti_hdu)

        out_timeline_hdu = createNewTimeline(timeline_hdu, t0, t1)
        if out_timeline_hdu is not None:
            ofd.append(out_timeline_hdu)

        updateKeywords(info, out_gti_hdu, t0, t1, nevents, ofd)

        ofd.writeto(filename)
        ofd.close()
        del ofd
        cosutil.printMsg("%s written" % filename)
        file_index += 1

def splitName(input):
    """Split the input name into rootname and suffix.

    Parameters
    ----------
    input: str
        Name of input file

    Returns
    -------
    tuple of two str
        The first string is the root name (up to "_corrtag"), and the
        second is everything past the root name.  For example, if
        input = "xyz_corrtag_a.fits", then the returned tuple would be
        ("xyz", "_corrtag_a.fits").
    """

    i = input.find("_corrtag")
    if i < 0:
        i = input.find(".fit")

    if i < 0:
        inroot = input
        suffix = ""
    else:
        inroot = input[:i]
        suffix = input[i:]

    return (inroot, suffix)

def getInfo(input, phdr, hdr):
    """Get header information.

    The input file name is included in the calling sequence just so it
    can be added to the info dictionary.

    Parameters
    ----------
    input: str
        Name of input file

    phdr: ``astropy.io.fits.header.Header`` object
        Primary header of input file

    hdr: ``astropy.io.fits.header.Header`` object
        EVENTS extension header of input file

    Returns
    -------
    info: dictionary
        Selected keywords and values from the input headers
    """

    info = {}

    info["input"] = input       # the name of the input file

    # This is a list of primary header keywords and default values.
    keylist = {
        "wavecorr":  "omit",
        "detector":  NOT_APPLICABLE,
        "segment":   NOT_APPLICABLE,
        "obstype":   NOT_APPLICABLE,
        "obsmode":   NOT_APPLICABLE,
        "exptype":   NOT_APPLICABLE,
        "opt_elem":  NOT_APPLICABLE}

    for key in keylist.keys():
        info[key] = phdr.get(key, default=keylist[key])

    # This is a list of extension header keywords and default values.
    # (also exptime; see below)
    keylist = {
        "expstart": -1.,
        "expend":   -1.}

    for key in keylist.keys():
        info[key] = hdr.get(key, default=keylist[key])
    exptime_key = cosutil.segmentSpecificKeyword("exptime", info["segment"])
    exptime_default = hdr.get("exptime", default=-1.)
    info["exptime"] = hdr.get(exptime_key, default=exptime_default)

    return info

def convertToSlices(time_col, starttime, increment, endtime, time_list):
    """Return a list of two-element tuples, giving time intervals.

    Parameters
    ----------
    time_col: array_like
        A copy of the TIME column from the input table

    starttime: float or None
        Time at beginning of first interval, or None if time_list was
        specified

    increment: float or None
        Length of each time interval, or None if time_list was specified

    endtime: float or None
        Time at end of last interval, or None if time_list was specified

    time_list: str or list
        The times of the beginning of each interval and the time of the
        end of the last interval; or time_list may already be the output
        format, a list of two-element tuples, in which case it will be
        returned unchanged

    Returns
    -------
    list of two-element tuples
        Each tuple in the list gives the start and end times (float) of an
        interval to be extracted from the input EVENTS table
    """

    if not increment and not time_list:
        raise RuntimeError("Must specify either increment or time_list.")

    if time_list and len(time_list) < 2:
        raise RuntimeError("time_list must have at least two elements.")

    if increment and time_list:
        cosutil.printWarning("Both increment and time_list were specified;"
                             " time_list will be used.")

    if time_list:

        if isinstance(time_list, str):
            time_list = convertToList(time_list)
        else:
            # If time_list is already a list of two-element tuples (or lists),
            # just return it.
            is_a_list_of_tuples = True          # default
            for value in time_list:
                if not isinstance(value, (list, tuple, np.ndarray)):
                    is_a_list_of_tuples = False
                    break
                if len(value) != 2:
                    is_a_list_of_tuples = False
                    break
            if is_a_list_of_tuples:
                return time_list

        # time_list is a list of times; extract the times into a list of
        # two-element tuples.
        new_time_list = []

        if isinstance(time_list[0], str):
            value = time_list[0].lower()
            if value == "start":
                t0 = 0.
            else:
                raise RuntimeError("First element of time_list is '%s'." %
                                   time_list[0])
        else:
            t0 = time_list[0]

        nelem = len(time_list)
        for i in range(1, nelem):
            if isinstance(time_list[i], str):
                value = time_list[i].lower()
                if value == "start":
                    t1 = 0.
                elif value == "stop" or value == "end":
                    t1 = time_col[-1]
                else:
                    raise RuntimeError(
                    "Don't understand time_list[%d] = '%s'." %
                        (i, time_list[i]))
            else:
                t1 = time_list[i]
            if t1 < t0:
                cosutil.printError("time_list = %s" % repr(time_list))
                raise RuntimeError("Values in time_list must be "
                                   "in increasing order.")
            elif t0 == t1:
                continue
            new_time_list.append((t0, t1))
            t0 = t1

    else:
        # Construct a list of two-element tuples from the start time and
        # increment, stopping at the end time.
        new_time_list = []

        if starttime is None:
            starttime = 0.
        t0 = starttime
        if endtime is None:
            endtime = time_col[-1]

        done = False
        while not done:
            if t0 >= endtime:
                done = True
            else:
                t1 = t0 + increment
                if t1 >= endtime:
                    t1 = endtime
                    done = True
                new_time_list.append((t0, t1))
                t0 = t1

    # Remove intervals that are outside the range of the data.
    trimmed_list = []
    for interval in new_time_list:
        if interval[1] > time_col[0] and interval[0] <= time_col[-1]:
            trimmed_list.append(interval)

    return trimmed_list

def convertToList(time_string):
    """Split a string of times on commas and/or blanks.

    Parameters
    ----------
    time_string: str
        String containing times, separated by blanks and/or commas; the
        times may be integers or floats, but "start", "stop" and "end" are
        also allowed

    Returns
    -------
    list of floats
        list of start and/or stop times of intervals, extracted from
        the input string
    """

    if time_string.find(",") >= 0:
        # split the string on commas, then strip whitespace
        words = []
        temp_words = time_string.split(",")
        for word in temp_words:
            words.extend(word.split())
    else:
        # split the string on blanks
        words = time_string.split()

    time_list = []
    for word in words:
        word = word.lower()
        if word == "start" or word == "stop" or word == "end":
            time_list.append(word)
        else:
            try:
                t_value = float(word)
            except ValueError:
                cosutil.printError("Values in time_list must be a number," \
                                   " or 'start', 'stop', or 'end'.")
                raise
            time_list.append(t_value)

    return time_list

def determineSlice(time_col, t0, t1):
    """Find the indices correspond to the start and end times of an interval.

    Parameters
    ----------
    time_col: array_like
        A copy of the TIME column from the input table

    t0: float
        Time at the beginning of an interval

    t1: float
        Time at the end of an interval

    Returns
    -------
    tuple of two integers
        The indices of a slice in time_col that will give the elements
        with times >= t0 and < t1
    """

    return ccos.range(time_col, t0, t1)

def constructOutputName(outroot, file_index, suffix):
    """Construct an output file name.

    For example, if outroot="abc", file_index=3, and suffix="_corrtag_a.fits",
    the output file name will be "abc_3_corrtag_a.fits".

    If a file with the constructed name already exists, however, a new name
    will be tried, a name with "_1" appended after the file_index.  If that
    file also already exists, "_2" will be tried instead of "_1", and so on
    until a name is found that doesn't exist.  For the above example, the
    first such name would be "abc_3_1_corrtag_a.fits".

    Parameters
    ----------
    outroot: str
        Root name with which to construct output file names

    file_index: int
        This number, preceded by "_", will be appended to outroot

    suffix: str
        All of the input file name that follows the root name

    Returns
    -------
    The name to use for the output file
    """

    upper_limit = 10000

    filename = "%s_%d%s" % (outroot, file_index, suffix)
    if os.access(filename, os.R_OK):
        save_filename = filename
        i = 1
        done = False
        while not done:
            filename = "%s_%d_%d%s" % (outroot, file_index, i, suffix)
            if os.access(filename, os.R_OK):
                i += 1
                if i > upper_limit:
                    raise RuntimeError("Output file already exists,"
                          " and upper limit of names has been exceeded.")
            else:
                done = True
        cosutil.printWarning("Output file %s already exists,"
                             " will use name %s instead." %
                             (save_filename, filename))

    return filename

def copyRows(data, ofd, i, j):
    """Copy rows from the input file to an output file.

    Parameters
    ----------
    data: ``astropy.io.fits.fitsrec.FITS_rec`` record array
        The data block of the EVENTS table extension

    ofd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        The data block of the EVENTS table extension

    i: int
        Number (zero indexed) of the first row to copy

    j: int
        One more than the number (zero indexed) of the last row to copy;
        that is, [i:j] is a slice
    """

    ofd[1].data = data[i:j]

def getGTI(ifd):
    """Find the most up-to-date GTI table in the input file.

    Parameters
    ----------
    ifd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        The fits file handle for the input file

    Returns
    -------
    ``astropy.io.fits.hdu.table.BinTableHDU`` object, or None
        The GTI extension from the input file, or None if there isn't one
    """

    # Find the GTI table with the largest value of EXTVER.
    last_extver = 0                     # initial value
    hdunum = 0
    for i in range(1, len(ifd)):
        hdu = ifd[i]
        extname = hdu.header.get("extname", "MISSING")
        if extname.upper() == "GTI":
            extver = hdu.header.get("extver", 1)
            if extver > last_extver:
                last_extver = extver
                hdunum = i

    if hdunum < 1:
        gti_hdu = None
    else:
        gti_hdu = ifd[hdunum]
    return gti_hdu

def createNewGTI(gti_hdu, t0, t1):
    """Create a GTI table for the output table.

    Parameters
    ----------
    gti_hdu: ``astropy.io.fits.hdu.table.BinTableHDU`` object, or None
        The GTI table from the input file (may be None)

    t0: float
        Time at the start of the interval

    t1: float
        Time at the end of the interval

    Returns
    -------
    ``astropy.io.fits.hdu.table.BinTableHDU`` object
        A GTI table to append to the output file
    """

    if gti_hdu is None or gti_hdu.data is None or len(gti_hdu.data) == 0:
        # No GTI table means the entire exposure is good.
        if gti_hdu is None or gti_hdu.data is None:
            col = []
            col.append(fits.Column(name="START", format="1D", unit="s"))
            col.append(fits.Column(name="STOP", format="1D", unit="s"))
            cd = fits.ColDefs(col)
        else:
            cd = gti_hdu.columns
        out_gti_hdu = fits.BinTableHDU.from_columns(cd, header=gti_hdu.header,
                                                    nrows=1)
        out_start_col = out_gti_hdu.data.field("start")
        out_stop_col = out_gti_hdu.data.field("stop")
        out_start_col[0] = t0
        out_stop_col[0] = t1
        return out_gti_hdu

    cd = gti_hdu.columns

    data = gti_hdu.data
    in_nrows = len(data)

    # columns in the GTI table from the input file
    start_col = data.field("start")
    stop_col = data.field("stop")

    gti = []                    # list of good (start, stop) intervals
    for i in range(in_nrows):
        start = start_col[i]
        stop = stop_col[i]
        if start >= t1 or stop <= t0:
            continue
        start = max(start, t0)
        stop = min(stop, t1)
        if (stop - start) <= 0.:
            continue
        gti.append((start, stop))

    if not gti:
        gti.append((0., 0.))
    out_nrows = len(gti)

    out_gti_hdu = fits.BinTableHDU.from_columns(cd, header=gti_hdu.header,
                                                nrows=out_nrows)
    out_start_col = out_gti_hdu.data.field("start")
    out_stop_col = out_gti_hdu.data.field("stop")
    for i in range(out_nrows):
        (start, stop) = gti[i]
        out_start_col[i] = start
        out_stop_col[i] = stop

    return out_gti_hdu

def getTimeline(ifd):
    """Get the TIMELINE extension (if there is one) from the input file.

    Parameters
    ----------
    ifd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        The fits file handle for the input file.

    Returns
    -------
    ``astropy.io.fits.hdu.table.BinTableHDU`` object, or None
        The TIMELINE extension from the input file, if there is one.
    """

    hdunum = 0
    for i in range(1, len(ifd)):
        hdu = ifd[i]
        extname = hdu.header.get("extname", "MISSING")
        if extname.upper() == "TIMELINE":
            hdunum = i
            break               # assume there's only one

    if hdunum < 1:
        timeline_hdu = None
    else:
        timeline_hdu = ifd[hdunum]
        # Note that this allows timeline_hdu.data to have zero length,
        # which will be checked for elsewhere and is OK.
        if timeline_hdu.data is None:
            timeline_hdu = None

    return timeline_hdu

def createNewTimeline(timeline_hdu, t0, t1):
    """Create a TIMELINE table for the output table.

    Parameters
    ----------
    timeline_hdu: ``astropy.io.fits.hdu.table.BinTableHDU`` object, or None
        The TIMELINE table from the input file (may be None).

    t0: float
        Time at the start of the interval.

    t1: float
        Time at the end of the interval.

    Returns
    -------
    ``astropy.io.fits.hdu.table.BinTableHDU`` object, or None
        A TIMELINE table to append to the output file, or None.  If
        there is no TIMELINE extension in the input file (indicated by
        timeline_hdu being None) or if the time increment is zero or
        negative, None will be returned.  Otherwise, the returned value
        will have the same columns as the input timeline_hdu, but the rows
        will be a subset of timeline_hdu.
    """

    if timeline_hdu is None:
        return None

    cd = timeline_hdu.columns

    in_data = timeline_hdu.data
    if in_data is None:
        in_nrows = 0
    else:
        in_nrows = len(in_data)

    if in_nrows > 0:
        time_col = in_data.field("time").astype(np.float64)
        # "ceil(t1) + 0.1" here is to ensure that the time range
        # (specifically i_end) actually includes all the relevant rows
        # of the input TIMELINE table.
        # This implicitly assumes that the time increment is one second.
        (i_start, i_end)= ccos.range(time_col, t0, math.ceil(t1) + 0.1)
        out_nrows = i_end - i_start
    else:
        out_nrows = 0

    out_timeline_hdu = fits.BinTableHDU.from_columns(cd,
                                                     header=timeline_hdu.header,
                                                     nrows=out_nrows)
    if in_nrows > 0:
        out_data = out_timeline_hdu.data
        i = i_start
        for j in range(out_nrows):
            out_data[j] = in_data[i]
            i += 1

    return out_timeline_hdu

def updateKeywords(info, out_gti_hdu, t0, t1, nevents, ofd):
    """Update keywords in an output file.

    This function adds two HISTORY records to the output primary header and
    updates EXPTIME, EXPTIMEA or EXPTIMEB, NEVENTS, NEVENTSA or NEVENTSB,
    EXPEND and EXPENDJ in the EVENTS extension header.

    Parameters
    ----------
    info: dictionary
        Keywords and values from the input header

    out_gti_hdu: ``astropy.io.fits.hdu.table.BinTableHDU`` object
        Header/data unit for the GTI table for the output file

    t0: float
        Time at the start of the interval

    t1: float
        Time at the end of the interval

    nevents: int
        Number of events in the output EVENTS table

    ofd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        The fits file handle for the output file
    """

    phdr = ofd[0].header
    hdr = ofd[1].header

    filename = os.path.basename(info["input"])          # just the file name
    phdr.add_history("Copied from %s" % filename)
    phdr.add_history("Time slice from input was %.3f to %.3f" % (t0, t1))

    if out_gti_hdu is None:
        exptime = t1 - t0
    else:
        start_col = out_gti_hdu.data.field("start")
        stop_col = out_gti_hdu.data.field("stop")
        exptime = 0.
        n = len(start_col)
        for i in range(n):
            exptime += (stop_col[i] - start_col[i])

    # Modified 2011 May 13 to update exptimea or exptimeb, depending on
    # segment.  Also update nevents and either neventsa or neventsb.
    hdr["exptime"] = exptime
    hdr["nevents"] = nevents
    if info["detector"] == "FUV":
        # first assign default values, so keywords for the "other" segment
        # will have the default
        hdr["exptimea"] = 0.
        hdr["exptimeb"] = 0.
        hdr["neventsa"] = 0
        hdr["neventsb"] = 0
        # "exptimea" or "exptimeb"
        exptime_key = cosutil.segmentSpecificKeyword("exptime",
                                                     info["segment"])
        hdr[exptime_key] = exptime
        nevents_key = "nevents" + info["segment"][-1].lower()
        hdr[nevents_key] = nevents

    expstart = info["expstart"]
    if expstart > 0.:
        expend = expstart + t1/SEC_PER_DAY
        hdr["expend"] = expend
        expend_j = expend + 2400000.5
        hdr["expendj"] = expend_j



import math
import os
import numpy as np
from astropy.io import fits
from . import cosutil
from . import dispersion
from . import orbit
from . import ccos
from .calcosparam import *       # parameter definitions

DIST_SUN  = 149597870.691       # 1 AU in km
RADIUS_EARTH = 6.367456e3       # geometric mean of equatorial and polar, km
SEC_PER_DAY = 86400.
MREF = 2451545. - 2400000.5     # MJD for 2000 Jan 1, 12h

TWOPI     = 2. * math.pi
ASECtoRAD = math.pi / (180. * 3600.)    # radians per arcsecond
DEGtoRAD  = math.pi / 180.              # radians per degree

# Wavelengths in Angstroms of airglow lines Lyman alpha, oxygen I, oxygen I.
# The values in the tuple are the wavelengths of the lines in the multiplet.
AIRGLOW_WAVELENGTHS = {"ly_alpha": (1215.67,),
                       "oi_1304": (1302.2, 1304.9, 1306.0),
                       "oi_1356": (1355.6, 1358.5),
                       "dark": None}

def createTimeline(input, fd, info, reffiles,
                   tl_time, shift1_vs_time,
                   time, xfull, yfull):
    """Create (or update) a timeline table.

    Parameters
    ----------
    input: str
        Name of input file (used only for creating spt file name).

    fd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        List of HDUs in corrtag file.

    info: dictionary
        Keywords and values.

    reffiles: dictionary
        Reference file keywords and names.

    tl_time: array_like
        The array of times corresponding to shift1_vs_time.

    shift1_vs_time: array_like
        The shifts in the dispersion direction at one-second intervals.
        or None if the current observation is a wavecal or if wavecal
        processing was not done.

    time: array_like
        The array of times of events in the corrtag table.

    xfull: array_like
        The array of fully corrected X positions, from the corrtag table.

    yfull: array_like
        The array of fully corrected Y positions, from the corrtag table.
    """

    # does the timeline extension already exist?
    try:
        hdu = fd[("timeline",1)]
        cosutil.printMsg("Update the TIMELINE extension.", VERBOSE)
    except KeyError:
        # timeline table doesn't exist; create it and append to corrtag file
        hdu = timelineHDU(len(tl_time), fd[1].header)
        cosutil.printMsg("Append a TIMELINE extension.", VERBOSE)
        fd.append(hdu)

    tl_data = hdu.data

    time_col = tl_data.field("time")
    if len(time_col) != len(tl_time):
        cosutil.printWarning("Number of rows in TIMELINE extension "
                "is %d, expected %d," % (len(time_col), len(tl_time)))
        cosutil.printContinuation("so TIMELINE table will not be updated.")
        return
    time_col[:] = tl_time.copy()

    if shift1_vs_time is None:
        median_shift1 = 0.
    else:
        shift1_col = tl_data.field("shift1")
        shift1_col[:] = shift1_vs_time.copy()
        index = shift1_vs_time.argsort()
        nelem = len(shift1_vs_time)
        mid_index = index[nelem // 2]
        median_shift1 = shift1_vs_time.item(mid_index)

    external_target = info["exptype"][:3] == "EXT"

    # columns for airglow count rates and dark counts
    ly_alpha_col = tl_data.field("ly_alpha")
    oi_1304_col = tl_data.field("oi_1304")
    oi_1356_col = tl_data.field("oi_1356")
    darkrate_col = tl_data.field("darkrate")
    # initialize to zero
    ly_alpha_col[:] = 0.
    oi_1304_col[:] = 0.
    oi_1356_col[:] = 0.
    darkrate_col[:] = 0.
    exptime = info["exptime"]
    if external_target and info["aperture"] in ["PSA", "BOA"] and \
       exptime > 0. and xfull is not None and yfull is not None:
        if len(tl_time) > 1:
            dt = tl_time[1] - tl_time[0]
        else:
            dt = 1.
        for key in ["ly_alpha", "oi_1304", "oi_1356", "dark"]:
            wl_airglow = AIRGLOW_WAVELENGTHS[key]
            region = findPixelRegion(info, reffiles["disptab"],
                                     reffiles["xtractab"],
                                     median_shift1, wl_airglow)
            if region is None:
                continue
            (y0, y1, x0, x1) = region
            if key == "ly_alpha" or key == "oi_1304":
                cosutil.printMsg("Airglow region for %s is "
                                 "X: %d to %d, Y: %d to %d" %
                                 (key.upper(), x0, x1+1, y0, y1+1), VERBOSE)
            # A value of 1 (True) in region_flags means the corresponding
            # event is within the area that includes the airglow line.
            region_flags = np.ones(len(xfull), dtype=np.bool_)
            region_flags = np.where(xfull > x1, False, region_flags)
            region_flags = np.where(xfull < x0,  False, region_flags)
            if isinstance(y0, (list, tuple)):
                # for the dark, there are two y regions
                region_flags = np.where(yfull > y1[1], False, region_flags)
                region_flags = np.where(yfull < y0[0],  False, region_flags)
                between = np.logical_and(yfull > y1[0], yfull < y0[1])
                region_flags = np.where(between,  False, region_flags)
                npixels = (x1 - x0) * (y1[0] - y0[0] + y1[1] - y0[1])
            else:
                region_flags = np.where(yfull > y1, False, region_flags)
                region_flags = np.where(yfull < y0,  False, region_flags)
                npixels = 1.
            region_flags = region_flags.astype(np.bool_)
            # scratch array for counts per second within each time bin
            temp = np.zeros(len(tl_time), dtype=np.float32)
            if time[-1] - time[0] < 1.:         # e.g. ACCUM data
                temp[:] = float(region_flags.sum(dtype=np.int32)) / exptime
            else:
                for i in range(len(tl_time)):
                    # jt0 and jt1 are indices in the TIME column, and therefore
                    # also in region_flags.
                    try:
                        (jt0, jt1) = ccos.range(time,
                                                tl_time[i], tl_time[i]+dt)
                        temp[i] = float(
                        (region_flags[jt0:jt1]).sum(dtype=np.int32)) / dt
                    except RuntimeError:
                        temp[i] = 0.
            if key == "ly_alpha":
                ly_alpha_col[:] = temp
            elif key == "oi_1304":
                oi_1304_col[:] = temp
            elif key == "oi_1356":
                oi_1356_col[:] = temp
            elif key == "dark":
                darkrate_col[:] = temp / npixels        # dark rate per pixel

    sptfile = makeSptFileName(input)
    if sptfile == "notfound":
        found_spt = False
    else:
        found_spt = os.access(sptfile, os.R_OK)
    if not found_spt:
        cosutil.printWarning("spt file not found, so TIMELINE extension "
                             "is incomplete")
        return

    orb = orbit.HSTOrbit(sptfile)

    sun_alt_col = tl_data.field("sun_alt")
    sun_zd_col = tl_data.field("sun_zd")
    long_col = tl_data.field("longitude")
    lat_col = tl_data.field("latitude")

    if external_target:
        ra_targ = info["ra_targ"] * DEGtoRAD
        dec_targ = info["dec_targ"] * DEGtoRAD
        rect_targ = sphToRect((1., ra_targ, dec_targ))      # unit vector
        target_alt_col = tl_data.field("target_alt")
        rv_col = tl_data.field("radial_vel")

    for i in range(len(tl_time)):
        mjd = np.float64(tl_time[i]) / SEC_PER_DAY + info["expstart"]
        (rect_hst, vel_hst) = orb.getPos(mjd)
        (r, ra_hst, dec_hst) = rectToSph(rect_hst)
        # Assume that we want geocentric latitude.  The difference from
        # astronomical latitude can be up to about 8.6 arcmin.
        lat_hst = dec_hst
        # Subtract the sidereal time at Greenwich to convert to longitude.
        long_hst = ra_hst - 2. * math.pi * gmst(mjd)
        if long_hst < 0.:
            long_hst += (2. * math.pi)
        long_col[i] = long_hst / DEGtoRAD
        lat_col[i] = lat_hst / DEGtoRAD
        rect_sun = eqSun(mjd)                   # equatorial coords of the Sun
        sun_alt_col[i] = computeAlt(rect_sun, rect_hst, parallax=True)
        sun_zd_col[i] = computeZD(rect_sun, rect_hst)
        if external_target:
            rv_col[i] = -dotProduct(rect_targ, vel_hst)
            target_alt_col[i] = computeAlt(rect_targ, rect_hst,
                                           parallax=False)

def makeSptFileName(input):
    """Construct the spt file name from the corrtag file name.

    Parameters
    ----------
    input: str
        Name of input file

    Returns
    --------
    str
        Name of support file, or "notfound".
    """
    pathname = os.path.dirname(input)
    filename = os.path.basename(input)
    rootname = filename.split("_")[0]
    sptfile = os.path.join(pathname, rootname + "_spt.fits")
    if not os.path.isfile:
        sptfile = "notfound"

    return sptfile

def sphToRect(sph):
    """Convert distance, RA and Dec to rectangular coordinates.

    Parameters
    ----------
    sph: array_like
        Distance (may be 1), longitude, latitude (angles in radians).

    Returns
    --------
    array_like
        Vector in rectangular coordinates.
    """

    (radius, longitude, latitude) = sph
    rect = np.zeros(3, dtype=np.float64)

    rect[0] = radius * math.cos(latitude) * math.cos(longitude)
    rect[1] = radius * math.cos(latitude) * math.sin(longitude)
    rect[2] = radius * math.sin(latitude)

    return rect

def rectToSph(rect):
    """Convert rectangular coordinates to RA and Dec.

    Parameters
    ----------
    rect: array_like
        Vector in rectangular coordinates.

    Returns
    --------
    array_like
        Distance, longitude, latitude (angles in radians).
    """

    # radius, longitude, latitude
    sph = np.zeros(3, dtype=np.float64)

    r2xy = rect[0] * rect[0] + rect[1] * rect[1]
    sph[0] = math.sqrt(r2xy + rect[2] * rect[2])
    rxy = math.sqrt(r2xy)
    sph[2] = math.atan2(rect[2], rxy)
    if r2xy > 0.:
        sph[1] = math.atan2(rect[1], rect[0])
        if sph[1] < 0.:
            sph[1] += (2. * math.pi)
    else:
        sph[1] = 0.

    return sph

def dotProduct(v1, v2):
    product = v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]
    return product.item()

def gmst(mjd):
    """Greenwich mean sidereal time at mjd.

    See Arthur N. Cox, Allen's Astrophysical Quantities, 4th ed., p 14.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The sidereal time at Greenwich, unit is fraction of a day.
    """

    d = mjd - MREF                      # days since MREF
    du = math.floor(mjd) - MREF
    Tu = du / 36525.
    # Greenwich mean sidereal time at 0h UT1, in seconds.
    GMST0 = 24110.54841 + 8640184.812866 * Tu \
                        + 0.093104 * Tu**2 - 6.2e-6 * Tu**3

    T = (mjd - MREF) / 36525.
    rprime = 1.002737909350795 + 5.9006e-11 * T - 5.9e-15 * T**2

    GMST = GMST0 / SEC_PER_DAY + (d - du) * rprime      # days
    t = GMST % 1.

    return t

def timelineHDU(nrows_timeline, hdr):
    """Create the TIMELINE HDU.

    Parameters
    ----------
    nrows_timeline: int
        Number of rows for the timeline table.

    hdr: ``astropy.io.fits.header.Header`` object
        EVENTS extension header.

    Returns
    -------
    ``astropy.io.fits.hdu.table.BinTableHDU`` object
        Header/data unit for a timeline table.
    """

    col = []
    col.append(fits.Column(name="TIME", format="1E",
                           unit="s", disp="F8.3"))
    col.append(fits.Column(name="LONGITUDE", format="1E",
                           unit="degree", disp="F10.6"))
    col.append(fits.Column(name="LATITUDE", format="1E",
                           unit="degree", disp="F10.6"))
    col.append(fits.Column(name="SUN_ALT", format="1E",
                           unit="degree", disp="F6.2"))
    col.append(fits.Column(name="SUN_ZD", format="1E",
                           unit="degree", disp="F6.2"))
    col.append(fits.Column(name="TARGET_ALT", format="1E",
                           unit="degree", disp="F6.2"))
    col.append(fits.Column(name="RADIAL_VEL", format="1E",
                           unit="km /s", disp="F7.5"))
    col.append(fits.Column(name="SHIFT1", format="1E",
                           unit="pixel", disp="F7.3"))
    col.append(fits.Column(name="LY_ALPHA", format="1E",
                           unit="count /s", disp="G15.6"))
    col.append(fits.Column(name="OI_1304", format="1E",
                           unit="count /s", disp="G15.6"))
    col.append(fits.Column(name="OI_1356", format="1E",
                           unit="count /s", disp="G15.6"))
    col.append(fits.Column(name="DARKRATE", format="1E",
                           unit="count /s /pixel", disp="G15.6"))
    cd = fits.ColDefs(col)

    #
    # Remove WCS keywords from table header
    newheader = cosutil.remove_WCS_keywords(hdr, cd)

    hdu = fits.BinTableHDU.from_columns(cd, header=newheader, nrows=nrows_timeline)

    hdu.header.set("extname", "TIMELINE", after="TFIELDS")      # xxx temp
    hdu.header.set("extver", 1, after="EXTNAME")        # xxx temporary
    # xxx hdu.header.insert("TFIELDS", ("extname", "TIMELINE"), after=True)
    # xxx hdu.header.insert("EXTNAME", ("extver", 1), after=True)

    return hdu

def computeAlt(rect, rect_hst, parallax=False):
    """Compute the altitude of an object, e.g. the Sun.

    This is based on the GC_sun_alt function, gc_sun_alt.cpp in OPUS.

    Parameters
    ----------
    rect: array like
        Rectangular, geocentric coordinates of an object.  If parallax due
        to the orbit of HST is significant, this vector should be in km
        (see parallax); otherwise, this may be a unit vector pointing
        toward the object.

    rect_hst: array like
        Rectangular, geocentric coordinates of HST, in km.

    parallax: boolean
        True if the object is close enough that we should correct for
        the parallax due to HST's offset from the center of the Earth.
        If True, rect should be in km; otherwise, rect may be a
        unit vector.

    Returns
    -------
    float
        Altitude of the object above the horizon, as seen by HST, in
        degrees.
    """

    if parallax:
        target = rect - rect_hst        # shift origin to HST
    else:
        target = rect
    rtarget = math.sqrt(dotProduct(target, target))
    # unit vector pointing from HST toward the object
    utarget = target / rtarget

    rhst = math.sqrt(dotProduct(rect_hst, rect_hst))
    # unit vector pointing from the center of the Earth toward HST
    uhst = rect_hst / rhst

    cz = dotProduct(uhst, utarget)      # cosine zenith distance
    zenith_dist = math.acos(cz)

    # The horizon is more than 90 degrees from the zenith; add this extra
    # angle.
    horizon_corr = math.acos(RADIUS_EARTH / rhst)
    altitude = math.pi / 2 + horizon_corr - zenith_dist

    return altitude / DEGtoRAD

def computeZD(rect_sun, rect_hst):
    """Compute the zenith distance (ignoring parallax) of the Sun.

    Parameters
    ----------
    rect_sun: array like
        Rectangular, geocentric coordinates of an object, in particular,
        the Sun.  The units are arbitrary because the vector will be
        normalized to 1.

    rect_hst: array like
        Rectangular, geocentric coordinates of HST.  The units are
        arbitrary because the vector will be normalized to 1.

    Returns
    -------
    float
        The angle between HST and the Sun, as seen from the center of
        the Earth, in degrees.
    """

    rsun = math.sqrt(dotProduct(rect_sun, rect_sun))
    usun = rect_sun / rsun

    rhst = math.sqrt(dotProduct(rect_hst, rect_hst))
    uhst = rect_hst / rhst

    # zenith_dist is not quite the zenith distance, because parallax
    # is not accounted for.
    cz = dotProduct(uhst, usun)
    zenith_dist = math.acos(cz)

    return zenith_dist / DEGtoRAD

def findPixelRegion(info, disptab, xtractab, median_shift1, wl_airglow):
    """Find the pixel region corresponding to an airglow line (or dark).

    Parameters
    ----------
    info: dictionary
        Keywords and values.

    disptab: str
        Name of reference table for dispersion solution.

    xtractab: str
        Name of extraction parameters reference table.

    median_shift1: float
        Median value in shift1_vs_time (in createTimeline()).  This is
        used when testing whether an airglow line is on the detector, and
        for shifting the region for the dark count rate.

    wl_airglow: tuple of floats, or None
        The elements of the tuple are the wavelengths of the airglow line
        or lines (i.e. in a multiplet), in Angstroms.
        If the region is for the dark count rate, wl_airglow will be None.

    Returns
    -------
    tuple
        (y0, y1, x0, x1)  x0 and x1 are floats; y0 and y1 are floats for an
        airglow line, or lists of two floats (each) for background regions.
        These are inclusive limits (pixels), not the elements of a slice.
        For an airglow line, y0 and y1 are the lower and upper limits of
        the airglow region, x0 and x1 are the left and right limits of the
        airglow region.
        For dark counts, x0 to x1 will cover the full width of the
        detector.  y0 and y1 are two-element lists; y0[0] and y0[1] are the
        lower limits of the two background regions, while y1[0] and y1[1]
        are the upper limits of the two background regions.
    """

    if info["obstype"] == "IMAGING":
        return None

    if info["detector"] == "FUV":
        segment = info["segment"]
        axis_length = FUV_X
        AIRGLOW_WIDTH = 184
        AIRGLOW_HEIGHT = 60
    else:
        segment = "NUVA"
        axis_length = NUV_X
        AIRGLOW_WIDTH = 60
        AIRGLOW_HEIGHT = 60

    x_width = float(AIRGLOW_WIDTH // 2)
    y_width = float(AIRGLOW_HEIGHT // 2)
    # fpoffset will be added to filter for the disptab.
    filter = {"opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"],
              "segment": segment,
              "aperture": info["aperture"]}
    xtract_info = cosutil.getTable(xtractab, filter)
    if xtract_info is None:
        cosutil.printWarning("No matching row in the XTRACTAB; filter is:")
        cosutil.printContinuation(str(filter))
        return None

    if wl_airglow is None:
        # Regions for dark count rate.
        x0 = 0. - median_shift1
        x1 = float(axis_length) - median_shift1
        b_bkg1 = xtract_info.field("b_bkg1")[0]
        b_bkg2 = xtract_info.field("b_bkg2")[0]
        if cosutil.findColumn(xtract_info, "b_hgt1"):
            bkg_height1 = xtract_info.field("b_hgt1")[0]
            bkg_height2 = xtract_info.field("b_hgt2")[0]
        else:
            bkg_height1 = xtract_info.field("bheight")[0]
            bkg_height2 = bkg_height1
        y0_low  = b_bkg1 - bkg_height1 // 2
        y0_high = b_bkg1 + bkg_height1 // 2
        y1_low  = b_bkg2 - bkg_height2 // 2
        y1_high = b_bkg2 + bkg_height2 // 2
        y0 = [y0_low, y1_low]
        y1 = [y0_high, y1_high]
        y0.sort()
        y1.sort()
    else:
        # Region for an airglow line.
        filter["fpoffset"] = info["fpoffset"]
        disp_rel = dispersion.Dispersion(disptab, filter)
        min_wl = min(wl_airglow)
        max_wl = max(wl_airglow)
        # First check whether the airglow line is off the detector.
        # NOTE that we assume that wavelength increases with x.
        wl_left_edge = disp_rel.evalDisp(-x_width - median_shift1)
        if max_wl < wl_left_edge:
            disp_rel.close()
            return None
        wl_right_edge = disp_rel.evalDisp(axis_length - 1. +
                                          x_width - median_shift1)
        if min_wl > wl_right_edge:
            disp_rel.close()
            return None
        # x_left and x_right are the pixel coordinates for the minimum
        # and maximum airglow wavelengths respectively.
        x_left = float(disp_rel.evalInvDisp(min_wl, tiny=1.e-8))
        x_right = float(disp_rel.evalInvDisp(max_wl, tiny=1.e-8))
        x0 = x_left - x_width
        x1 = x_right + x_width
        disp_rel.close()
        slope = xtract_info.field("slope").item(0)
        b_spec = xtract_info.field("b_spec").item(0)
        y = b_spec + slope * (x_left + x_right) / 2.
        y0 = y - y_width
        y1 = y + y_width

    return (y0, y1, x0, x1)

def eqSun(mjd):
    """Compute the equatorial coordinates of the Sun.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    rect_sun: array_like
        Three-element vector containing the equatorial rectangular
        coordinates of the Sun, in km.
    """

    ecl_sun = eclSun(mjd)               # ecliptic coordinates of the Sun
    rect_sun = eclToEq(ecl_sun, mjd)    # convert to equatorial coordinates
    rect_sun *= DIST_SUN                # convert distance from AU to km

    return rect_sun

def eclSun(mjd):
    """Compute the ecliptic coordinates of the Sun.

    This is based on Pulkinnen & VanFlandern, ApJ Suppl Series, vol. 41,
    Nov. 1979, p. 391-41.  The accuracy is of order a minute of arc (but
    often much better), as seen from Earth.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    eclcoord: array_like
        Three-element vector containing the ecliptic rectangular
        coordinates of the Sun at mjd, in astronomical units.
    """

    eclcoord = np.zeros(3, dtype=np.float64)

    tJcent = (mjd - MREF) / 36525. + 1.         # Julian centuries

    ls = lsun(mjd)
    gs = gsun(mjd)
    lm = lmoon(mjd)
    gv2 = gvenus(mjd)
    gm4 = gmars(mjd)
    gj5 = gjupiter(mjd)

    sings = math.sin(gs)
    cosgs = math.cos(gs)
    sin2gs = 2. * sings * cosgs
    cos2gs = cosgs * cosgs - sings * sings

    ecllong = ls + (6910. * sings + 72. * sin2gs - 17. * tJcent * sings \
                     - 7. * math.cos(gs - gj5) \
                     + 6. * math.sin(lm - ls) \
                     + 5. * math.sin(4.*gs - 8.*gm4 + 3.*gj5) \
                     - 5. * math.cos(2.*gs - 2.*gv2) \
                     - 4. * math.sin(gs - gv2) \
                     + 4. * math.cos(4.*gs - 8.*gm4 + 3.*gj5) \
                     + 3. * math.sin(2.*gs - 2.*gv2) \
                     - 3. * math.sin(gj5) \
                     - 3. * math.sin(2.*gs - 2.*gj5)) * ASECtoRAD

    dist = 1.00014 - 0.01675 * cosgs - 0.00014 * cos2gs

    eclcoord[0] = dist * math.cos(ecllong)
    eclcoord[1] = dist * math.sin(ecllong)
    eclcoord[2] = 0.

    return eclcoord

def eclToEq(eclcoord, mjd):
    """Convert from ecliptic coordinates to equatorial coordinates.

    Parameters
    ----------
    eclcoord: array_like
        Three-element vector containing ecliptic rectangular coordinates.

    mjd: float
        The Modified Julian Date.

    Returns
    -------
    eqcoord: array_like
        Three-element vector containing the equatorial rectangular
        coordinates corresponding to eclcoord at mjd.
    """

    eqcoord = np.zeros(3, dtype=np.float64)

    tJcent = (mjd - MREF) / 36525. + 1.         # Julian centuries

    omega = momega(mjd)         # longitude of the ascending node of the moon

    eps = (84428. - 47. * tJcent + 9. * math.cos(omega)) * ASECtoRAD
    coseps = math.cos(eps)
    sineps = math.sin(eps)

    # Rotate around x-axis by the obliquity of the ecliptic.
    eqcoord[0] = eclcoord[0]
    eqcoord[1] = coseps * eclcoord[1] - sineps * eclcoord[2]
    eqcoord[2] = sineps * eclcoord[1] + coseps * eclcoord[2]

    return eqcoord

def lsun(mjd):
    """Compute the mean longitude of the Sun.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The mean longitude of the Sun at mjd, in radians.
    """

    return TWOPI * (0.779072 + 0.00273790931 * (mjd - MREF))

def gsun(mjd):
    """Compute the mean anomaly of the Sun.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The mean anomaly of the Sun at mjd, in radians.
    """

    return TWOPI * (0.993126 + 0.00273777850 * (mjd - MREF))

def lmoon(mjd):
    """Compute the mean longitude of the Moon.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The mean longitude of the Moon at mjd, in radians.
    """

    return TWOPI * (0.606434 + 0.03660110129 * (mjd - MREF))

def momega(mjd):
    """Compute L_moon - F_moon for the Moon.

    L_moon is the mean longitude of the Moon, and F_moon is the argument
    of latitude of the Moon.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        L_moon - F_moon at mjd, in radians.
    """

    return TWOPI * (0.347343 - 0.00014709391 * (mjd - MREF))

def gvenus(mjd):
    """Compute the mean anomaly for Venus.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The mean anomaly for Venus at mjd, in radians.
    """

    return TWOPI * (0.140023 + 0.00445036173 * (mjd - MREF))

def gmars(mjd):
    """Compute the mean anomaly for Mars.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The mean anomaly for Mars at mjd, in radians.
    """

    return TWOPI * (0.053856 + 0.00145561327 * (mjd - MREF))

def gjupiter(mjd):
    """Compute the mean anomaly for Jupiter.

    Parameters
    ----------
    mjd: float
        The Modified Julian Date.

    Returns
    -------
    float
        The mean anomaly for Jupiter at mjd, in radians.
    """

    return TWOPI * (0.056531 + 0.00023080893 * (mjd - MREF))



from calcos import airglow
from generate_tempfiles import create_disptab_file
import os


def test_find_airglow_limits(tmp_path):
    """
    unit test for find_airglow_limits()
    test ran
    - By providing certain values as dict to be used as filter for finding the dispersion
    - testing for both FUV segments
    - creating a temporary disptab ref file.
    - testing for 5 airglow lines
    - calculating the expected pixel numbers by following the math involved in the actual file
        and referring to the values in the ref file we can get the values upto a descent decimal points.

    Returns
    -------
    pass if expected == actual or fail if not.

    """
    # Setup
    inf = {"obstype": "SPECTROSCOPIC", "cenwave": 1055, "aperture": "PSA", "detector": "FUV",
           "opt_elem": "G130M", "segment": "FUVA"}
    seg = ["FUVA", "FUVB"]
    disptab = create_disptab_file(str(tmp_path / '49g17153l_disp.fits'))
    airglow_lines = ["Lyman_alpha", "N_I_1200", "O_I_1304", "O_I_1356", "N_I_1134"]
    actual_pxl = [
        [], [], (15421.504705213156, 15738.02214190493), (8853.838672375898, 9135.702216258482)]
    # Test
    test_pxl = [[], []]
    # only works for FUV
    for segment in seg:
        for line in airglow_lines:
            limits = airglow.findAirglowLimits(inf, segment, disptab, line)
            if limits is not None:
                x, y = limits
                test_pxl.append((x, y))
    # Verify
    for i in range(len(actual_pxl)):
        assert actual_pxl[i] == test_pxl[i]



from calcos import shiftfile

def create_shift_file(filename):
    # Create the shift file for use in tests
    with open(filename, "w") as file:
        file.write("#dataset\tfpoffset\tflash #\tstripe\tshift1\tshift2\n")
        for i in range(10):
            if i % 3 == 0:
                file.write("{}\t{}\t{}\t{}\t{}\t{}\n".format("abc123def", "any", "1", "NUVA", "45.234435", "7"))
            elif i % 5 == 0:
                file.write("{}\t{}\t{}\t{}\t{}\t{}\n".format("ghi456jkl", "any", "2", "NUVB", "34.543453", "7"))
            elif i % 6 == 0 or i % 8 == 0:
                file.write("{}\t{}\t{}\t{}\t{}\t{}\n".format("ghi456jkl", "any", "2", "FUVA", "19.543453", "5"))
            elif i == 9:
                file.write("{}\t{}\t{}\t{}\t{}\t{}\n".format("mno789pqr", "any", "2", "FUVB", "52.723453", "6"))
            else:
                file.write("{}\t{}\t{}\t{}\t{}\t{}\n".format("mno789pqr", "any", "1", "NUVC", "-34.543453", "7"))

    return

def test_shift_file(tmp_path):
    shift_file = str(tmp_path / "shift_file.txt")
    create_shift_file(shift_file)
    # Test
    ob = shiftfile.ShiftFile(shift_file, 'abc123def', 'any')
    # Verify
    assert len(ob.user_shift_dict) > 0


def test_get_shifts(tmp_path):
    # Setup
    shift_file = str(tmp_path / "shift_file.txt")
    create_shift_file(shift_file)
    ob1 = shiftfile.ShiftFile(shift_file, 'ghi456jkl', 'any')
    ob2 = shiftfile.ShiftFile(shift_file, 'abc123def', 'any')
    keys = [('any', 'nuva'), ('any', 'nuvb'), (2, 'nuvc'), ('any', 'any'), ('any', 'fuva'), ('any', 'fuvb')]
    expected_values1 = [((None, None), 0), ((34.543453, 7.0), 1), ((None, None), 0), ((19.543453, 5.0), 2), ((19.543453, 5.0), 1), ((None, None), 0)]
    expected_values2 = [((45.234435, 7.0), 1), ((None, None), 0), ((None, None), 0), ((45.234435, 7.0), 1), ((None, None), 0), ((None, None), 0)]
    # Test
    test_values1 = []
    test_values2 = []
    for key in keys:
        test_values1.append(shiftfile.ShiftFile.getShifts(ob1, key))
    for key in keys:
        test_values2.append(shiftfile.ShiftFile.getShifts(ob2, key))
    # Verify
    for i in range(len(expected_values1)):
        assert expected_values1[i] == test_values1[i]
        assert expected_values2[i] == test_values2[i]

